import html2text
import mailparser


class Email:
    def __init__(self, raw):
        self.raw = raw
        self.email = mailparser.parse_from_string(self.raw)

        self.message_id: str = self.email.message_id
        self.subject: str = self.email.subject or ''
        self.body: str = self.email.body

        self.recipients: list[str]
        self.recipient_names: list[str]
        self.senders: str
        self.sender_names: str
        self.cc: list[str]
        self.cc_names: list[str]
        self.bcc: list[str]
        self.bcc_names: list[str]
        self.reply_to: list[str]
        self.reply_to_names: list[str]

        def zip_if_nonempty(arr):
            return zip(*arr) if arr else ([], [])

        self.recipient_names, self.recipients = zip_if_nonempty(self.email.to)
        self.sender_names, self.senders = zip_if_nonempty(self.email.from_)
        self.cc_names, self.cc = zip_if_nonempty(self.email.cc)
        self.bcc_names, self.bcc = zip_if_nonempty(self.email.bcc)
        self.reply_to_names, self.reply_to = zip_if_nonempty(self.email.reply_to)

        self.date = self.email.date
        self.text_plain: str = '\n'.join(self.email.text_plain)
        self.text_html: str = '\n'.join(self.email.text_html)

        self.attachments = self.email.attachments
        self.headers = self.email.headers

        h = html2text.HTML2Text()
        h.ignore_images = True
        self.text_markdown: str = h.handle(self.text_html)

    @classmethod
    def from_file(cls, filename):
        with open(filename) as f:
            return cls(f.read())



from emailrouter.email import Email
from emailrouter.filters import FilterRegistry
from emailrouter.handlers import HandlerRegistry
from emailrouter.routing import Route, Router
from emailrouter.utils import ArgumentMixin, Base, InvalidConfigException, Registry, load_module, load_module_from_file



import os
import types


def load_module(name, code, filename=None):
    mod = types.ModuleType(name)
    if filename is not None:
        mod.__file__ = filename
    exec(compile(code, filename or '<string>', 'exec'), mod.__dict__)
    return mod


def load_module_from_file(filename):
    path, name = os.path.split(filename)
    name, ext = os.path.splitext(name)

    with open(filename) as f:
        return load_module(name, f.read(), os.path.abspath(filename))


class InvalidConfigException(Exception):
    pass


class RegistryInterface:
    def __init__(self, registry):
        self._registry: 'Registry' = registry
        self.named_objects = {}

    def create_object(self, data):
        return self._registry.create_object(data, self) or self.named_objects[data['name']]

    def register_named_object(self, name, data):
        self.named_objects[name] = self.create_object(data)


class Registry:
    def __init__(self, named_type='named'):
        self.classes = {named_type: None}
        self.default_class = None

    def register_class(self, name, default=False):
        def register(clazz):
            clazz.name = name
            assert name not in self.classes
            self.classes[name] = clazz
            if default:
                assert self.default_class is None
                self.default_class = clazz
            return clazz

        return register

    def create_object(self, data, registry):
        if data is None:
            if self.default_class is not None:
                clazz = self.default_class
            else:
                raise InvalidConfigException('No default class available and no type specified.')
        else:
            t = data['type']
            if t not in self.classes:
                raise InvalidConfigException(f'Type "{t}" not found.')
            clazz = self.classes[t]

        return clazz(data, registry) if clazz is not None else None

    def __call__(self):
        return RegistryInterface(self)


class Base:
    name: str

    def __init__(self, data, registry):
        if data is not None and 'type' in data:
            assert data['type'] == self.name
        self.data = data
        self.registry: RegistryInterface = registry

    def __call__(self, email):
        raise NotImplementedError()

    def __str__(self):
        return self.name


class ArgumentMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.args = self.data.get('args', ())
        self.kwargs = self.data.get('kwargs', {})



import re

from emailrouter.utils import ArgumentMixin, Base as Filter, Registry, load_module_from_file


FilterRegistry = Registry()
register_filter = FilterRegistry.register_class


@register_filter('python')
class PythonFilter(ArgumentMixin, Filter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.call_func = load_module_from_file(self.data['file']).check

    def __call__(self, email):
        return self.call_func(email, *self.args, **self.kwargs)


class NonLeafFilter(Filter):
    pass


class MultipleFilter(NonLeafFilter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subfilters = [self.registry.create_object(f) for f in self.data['conditions']]


class SingleFilter(NonLeafFilter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subfilter = self.registry.create_object(self.data['condition'])


class LeafFilter(Filter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.field = self.process_field(self.data['field'])
        self.value = self.process_value(self.data['value'])

    def field_data(self, data):
        if self.field is None:
            return data
        return getattr(data, self.field)

    @classmethod
    def process_value(cls, value):
        return value

    @classmethod
    def process_field(cls, field):
        assert field in (
            None,
            'message_id',
            'bcc',
            'bcc_names',
            'body',
            'cc',
            'cc_names',
            'recipient_names',
            'recipients',
            'reply_to',
            'reply_to_names',
            'sender_names',
            'senders',
            'subject',
        )
        return field


class CaseInsensitiveMixin:
    @classmethod
    def process_value(cls, value):
        return super().process_value(value).lower()

    def field_data(self, email):
        return super().field_data(email).lower()


@register_filter('true', default=True)
class TrueFilter(Filter):
    def __call__(self, email):
        return True


@register_filter('false')
class FalseFilter(Filter):
    def __call__(self, email):
        return False


@register_filter('any')
class AnyFilter(MultipleFilter):
    def __call__(self, email):
        return any(f(email) for f in self.subfilters)


@register_filter('all')
class AllFilter(MultipleFilter):
    def __call__(self, email):
        return all(f(email) for f in self.subfilters)


@register_filter('not')
class NotFilter(SingleFilter):
    def __call__(self, email):
        return not self.subfilter(email)


@register_filter('equal')
class EqualFilter(LeafFilter):
    def __call__(self, email):
        return self.field_data(email) == self.value


@register_filter('iequal')
class IEqualFilter(CaseInsensitiveMixin, EqualFilter):
    pass


@register_filter('regex')
class RegexFilter(LeafFilter):
    flags = 0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.re_expr = re.compile(self.value, flags=self.flags)

    def __call__(self, email):
        return self.re_expr.match(self.field_data(email)) is not None


@register_filter('iregex')
class IRegexFilter(RegexFilter):
    flags = re.IGNORECASE


@register_filter('substring')
class SubstringFilter(LeafFilter):
    def __call__(self, email):
        return self.value in self.field_data(email)


@register_filter('isubstring')
class ISubstringFilter(CaseInsensitiveMixin, SubstringFilter):
    pass


@register_filter('startswith')
class StartsWithFilter(LeafFilter):
    def __call__(self, email):
        return self.field_data(email).startswith(self.value)


@register_filter('istartswith')
class IStartsWithFilter(CaseInsensitiveMixin, StartsWithFilter):
    pass


@register_filter('endswith')
class EndsWithFilter(LeafFilter):
    def __call__(self, email):
        return self.field_data(email).endswith(self.value)


@register_filter('iendswith')
class IEndsWithFilter(CaseInsensitiveMixin, EndsWithFilter):
    pass


@register_filter('length')
class LengthFilter(LeafFilter):
    def __call__(self, email):
        return len(self.field_data(email)) == self.value


@register_filter('listany')
class ListAnyFilter(LeafFilter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subfilter = self.registry.create_object({
            'type': self.data['condition'],
            'field': None,
            'value': self.data['value'],
        })

    def __call__(self, email):
        return any(self.subfilter(data) for data in self.field_data(email))


@register_filter('listall')
class ListAllFilter(LeafFilter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subfilter = self.registry.create_object({
            'type': self.data['condition'],
            'field': None,
            'value': self.data['value'],
        })

    def __call__(self, email):
        return all(self.subfilter(data) for data in self.field_data(email))



import imaplib
import json
import urllib.parse
from urllib import request

from emailrouter.utils import ArgumentMixin, Base as Handler, Registry, load_module_from_file

HandlerRegistry = Registry()
register_handler = HandlerRegistry.register_class


@register_handler('python')
class PythonHandler(ArgumentMixin, Handler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.call_func = load_module_from_file(self.data['file']).handle

    def __call__(self, email):
        return self.call_func(email, *self.args, **self.kwargs)


@register_handler('placeholder', default=True)
class PlaceholderHandler(Handler):
    def __call__(self, email):
        pass


@register_handler('imap')
class IMAPHandler(ArgumentMixin, Handler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        ssl = self.kwargs['ssl']
        url = urllib.parse.urlparse(self.kwargs['url'])
        default_port = 993 if ssl else 143

        self.mailbox = self.kwargs.get('mailbox')
        self.flags = self.kwargs.get('flags')
        self.imap = imaplib.IMAP4_SSL if ssl else imaplib.IMAP4
        self.credentials = {
            'host': url.hostname,
            'port': url.port or default_port,
            'username': url.username,
            'password': url.password,
        }

    def __call__(self, email):
        conn = self.imap(host=self.credentials['host'], port=self.credentials['port'])
        conn.login(self.credentials['username'], self.credentials['password'])
        conn.append(self.mailbox, self.flags, None, email.raw.encode('utf-8'))
        conn.logout()


class DiscordWebhook:
    MESSAGE_LIMIT = 3000

    def __init__(self, webhook_url):
        self.webhook_url = webhook_url

    def message(self, message):
        req = request.Request(
            self.webhook_url,
            data=json.dumps(message).encode('utf-8'),
            headers={'User-Agent': 'Python/3', 'Content-Type': 'application/json'},
        )
        request.urlopen(req)

    @staticmethod
    def escape(text):
        escape_chars = ('*', '_', '~', '`', '|')
        for ch in escape_chars:
            text = text.replace(ch, '\\' + ch)
        return text


@register_handler('discord')
class DiscordHandler(ArgumentMixin, Handler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.webhook = DiscordWebhook(self.kwargs['webhook_url'])
        self.payload = self.kwargs.get('payload', {})
        self.colour = self.kwargs.get('color', self.kwargs.get('colour', 0xeeeeee))

    def __call__(self, email):
        subject = self.webhook.escape(email.subject)
        message = self.webhook.escape(email.text_plain) or email.text_markdown

        fields = []
        for name, value in (
            ('From', ', '.join(email.senders)),
            ('To', ', '.join(email.recipients)),
            ('CC', ', '.join(email.cc)),
            ('BCC', ', '.join(email.bcc)),
            ('Reply To', ', '.join(email.reply_to)),
        ):
            if value:
                fields.append({
                    'name': name,
                    'value': value,
                })

        self.webhook.message({
            **self.payload,
            'embeds': [
                {
                    'title': subject,
                    'color': self.colour,
                    'timestamp': email.date.isoformat(),
                    'fields': fields,
                },
                {
                    'title': subject,
                    'description': message[:self.webhook.MESSAGE_LIMIT],
                    'color': self.colour,
                    'timestamp': email.date.isoformat(),
                },
            ],
        })



import unittest

from emailrouter import FilterRegistry, InvalidConfigException


class TestFilters(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.registry = FilterRegistry()

    def test_true(self):
        f = self.registry.create_object({
            'type': 'true',
        })
        self.assertTrue(f(' '))

        f = self.registry.create_object(None)
        self.assertTrue(f(''))

    def test_false(self):
        f = self.registry.create_object({
            'type': 'false',
        })
        self.assertFalse(f(' '))

    def test_equal(self):
        f = self.registry.create_object({
            'type': 'equal',
            'field': None,
            'value': 'value',
        })
        self.assertTrue(f('value'))
        self.assertFalse(f('VALUe'))
        self.assertFalse(f('value2'))

    def test_iequal(self):
        f = self.registry.create_object({
            'type': 'iequal',
            'field': None,
            'value': 'value',
        })
        self.assertTrue(f('value'))
        self.assertTrue(f('ValuE'))
        self.assertFalse(f('vvalue'))

    def test_regex(self):
        f = self.registry.create_object({
            'type': 'regex',
            'field': None,
            'value': r'a.*@gmail\.com',
        })
        self.assertTrue(f('a.@gmail.com'))
        self.assertTrue(f('abbbb@gmail.com'))
        self.assertTrue(f('aBBB@gmail.com'))
        self.assertFalse(f('aBBB@GMAIL.com'))
        self.assertFalse(f('ba@gmail.com'))
        self.assertFalse(f('a@gmail\\.com'))

    def test_iregex(self):
        f = self.registry.create_object({
            'type': 'iregex',
            'field': None,
            'value': r'a.*@gmail\.com',
        })
        self.assertTrue(f('abbbb@gmail.com'))
        self.assertTrue(f('aBBB@gmail.com'))
        self.assertTrue(f('aBBB@GMAIL.com'))
        self.assertFalse(f('ba@gmail.com'))

    def test_substring(self):
        f = self.registry.create_object({
            'type': 'substring',
            'field': None,
            'value': 'abcdef',
        })
        self.assertTrue(f('abcdef'))
        self.assertTrue(f('wwabcdefgg'))
        self.assertTrue(f('aaaabcdeffff'))
        self.assertFalse(f('aaaabcdeFfff'))
        self.assertFalse(f('wwabcdegg'))

    def test_isubstring(self):
        f = self.registry.create_object({
            'type': 'isubstring',
            'field': None,
            'value': 'abcdEF',
        })
        self.assertTrue(f('abcdef'))
        self.assertTrue(f('aaaabcdeffff'))
        self.assertTrue(f('aaaabcdeFfff'))
        self.assertFalse(f('wwabcdegg'))

    def test_startswith(self):
        f = self.registry.create_object({
            'type': 'startswith',
            'field': None,
            'value': 'ab',
        })
        self.assertTrue(f('abcd'))
        self.assertFalse(f('aBab'))

    def test_istartswith(self):
        f = self.registry.create_object({
            'type': 'istartswith',
            'field': None,
            'value': 'ab',
        })
        self.assertTrue(f('abcd'))
        self.assertTrue(f('aBab'))
        self.assertFalse(f('aab'))

    def test_endswith(self):
        f = self.registry.create_object({
            'type': 'endswith',
            'field': None,
            'value': 'cd',
        })
        self.assertTrue(f('abcd'))
        self.assertFalse(f('cdcD'))

    def test_iendswith(self):
        f = self.registry.create_object({
            'type': 'iendswith',
            'field': None,
            'value': 'cd',
        })
        self.assertTrue(f('abcd'))
        self.assertTrue(f('cdcD'))
        self.assertFalse(f('cdcddc'))

    def test_length(self):
        f = self.registry.create_object({
            'type': 'length',
            'field': None,
            'value': 2,
        })
        self.assertTrue(f('aa'))
        self.assertTrue(f('=]'))
        self.assertFalse(f('"'))

    def test_listany(self):
        f = self.registry.create_object({
            'type': 'listany',
            'field': None,
            'value': 'cd',
            'condition': 'iequal',
        })
        self.assertTrue(f(['a', 'b', 'cd']))
        self.assertTrue(f(['a', 'cD', 'b']))
        self.assertFalse(f(['a', 'cdd', 'b']))

    def test_listall(self):
        f = self.registry.create_object({
            'type': 'listall',
            'field': None,
            'value': '@hotmail.com',
            'condition': 'substring',
        })
        self.assertTrue(f(['@hotmail.com', '@hotmail.comaa', 'aa@hotmail.com.']))
        self.assertTrue(f(['@hotmail.com@hotmail.com']))
        self.assertFalse(f(['@hotmail.com', 'cdd']))

    def test_any(self):
        f = self.registry.create_object({
            'type': 'any',
            'conditions': [
                {
                    'type': 'equal',
                    'field': None,
                    'value': 'p',
                },
                {
                    'type': 'regex',
                    'field': None,
                    'value': '.*aa(.*)',
                },
            ],
        })
        self.assertTrue(f('p'))
        self.assertTrue(f('aap'))
        self.assertTrue(f('paoskaaa'))
        self.assertFalse(f('pppp'))

    def test_all(self):
        f = self.registry.create_object({
            'type': 'all',
            'conditions': [
                {
                    'type': 'istartswith',
                    'field': None,
                    'value': 'p',
                },
                {
                    'type': 'iregex',
                    'field': None,
                    'value': '.*aa(.*)',
                },
            ],
        })
        self.assertTrue(f('paoskaaa'))
        self.assertFalse(f('p'))
        self.assertFalse(f('aap'))
        self.assertFalse(f('pppp'))

    def test_not(self):
        f = self.registry.create_object({
            'type': 'not',
            'condition': {
                'type': 'equal',
                'field': None,
                'value': 'llll',
            },
        })
        self.assertTrue(f('lll'))
        self.assertFalse(f('llll'))

    def test_invalid(self):
        with self.assertRaises(InvalidConfigException):
            self.registry.create_object({'type': 'aaa'})



import logging

import yaml

from emailrouter.filters import FilterRegistry
from emailrouter.handlers import HandlerRegistry


logger = logging.getLogger(__name__)


class Route:
    def __init__(self, data, router):
        self.data = data
        self.router = router

        self.name = self.data['name']
        self.filter = router.filter_registry.create_object(self.data['condition'])
        self.handlers = list(map(router.handler_registry.create_object, self.data['handlers']))
        self.propagate = self.data.get('propagate', True)

    def __call__(self, email):
        if not self.filter(email):
            return False
        for handler in self.handlers:
            handler(email)

        if self.propagate:
            return True
        return None

    def __str__(self):
        return self.name


class Router:
    def __init__(self, config):
        self.filter_registry = FilterRegistry()
        self.handler_registry = HandlerRegistry()

        if 'handlers' in config:
            for name, data in config['handlers'].items():
                self.handler_registry.register_named_object(name, data)
        if 'filters' in config:
            for name, data in config['filters'].items():
                self.filter_registry.register_named_object(name, data)
        self.routes = [Route(
            data,
            router=self,
        ) for data in config['routes']]

    def execute(self, email):
        processed_routes = []
        for route in self.routes:
            logger.debug('Running route "%s".', route)
            route_result = route(email)
            if route_result is not False:
                logger.debug('Route "%s" ran.', route)
                processed_routes.append(str(route))
            else:
                logger.debug('Route "%s" skipped.', route)
            if route_result is None:
                break
        logger.info('Ran %d route(s) for email %s.', len(processed_routes), email.message_id)
        return processed_routes

    @classmethod
    def from_yaml(cls, yaml_data):
        return cls(yaml.safe_load(yaml_data))

    @classmethod
    def from_yaml_file(cls, yaml_file):
        with open(yaml_file) as f:
            return cls.from_yaml(f.read())



import logging
import os

import yaml

from emailrouter import Email, Router


logger = logging.getLogger(__name__)


class Tester:
    def __init__(self, test_root):
        self.test_root = test_root

    def run_test(self, test_name):
        logger.info('Running test %s.', test_name)
        with open('test.yml') as f:
            test_config = yaml.safe_load(f.read())
        router = Router.from_yaml_file(test_config['config'])

        success = 0
        total = len(test_config['emails'])
        for email in test_config['emails']:
            em = Email.from_file(email['file'])
            actual = router.execute(em)
            if actual == email['expected']:
                success += 1
            else:
                logging.error(
                    'Test %s - %s failed. Expected: %s. Actual: %s.',
                    test_name,
                    email['name'],
                    email['expected'],
                    actual,
                )
        logging.info('Test %s results: %d/%d.', test_name, success, total)
        return success == total

    def test_all(self):
        pwd = os.getcwd()

        success = 0
        total = 0
        for d in os.listdir(self.test_root):
            path = os.path.join(self.test_root, d)
            if not os.path.isdir(path) or 'test.yml' not in os.listdir(path):
                continue
            os.chdir(path)
            try:
                success += self.run_test(d)
            except:  # noqa
                logging.exception('Test %s crashed.', d)
            total += 1
            os.chdir(pwd)
        logging.info('Tests: %d/%d', success, total)
        return success == total



import unittest
from email.message import EmailMessage

from emailrouter import Email, Router


def handle(email):
    raise ValueError()


class TestRouter(unittest.TestCase):
    def test_basic(self):
        r = Router({
            'handlers': {
                'handler': {
                    'type': 'imap',
                    'kwargs': {
                        'ssl': True,
                        'url': 'imap://a:a@127.0.0.1',
                        'mailbox': 'Trash',
                    },
                },
            },
            'filters': {
                'name': {
                    'type': 'true',
                },
            },
            'routes': [
                {
                    'name': 'main',
                    'handlers': [
                        {
                            'type': 'python',
                            'file': __file__,
                        },
                        {
                            'type': 'named',
                            'name': 'handler',
                        },
                    ],
                    'condition': {
                        'type': 'all',
                        'conditions': [
                            {
                                'type': 'named',
                                'name': 'name',
                            },
                            {
                                'type': 'equal',
                                'field': 'subject',
                                'value': 'RE:',
                            },
                            {
                                'type': 'length',
                                'field': 'subject',
                                'value': 3,
                            },
                        ],
                    },
                },
            ],
        })
        msg = EmailMessage()
        msg['Subject'] = 'RE:'
        msg['To'] = 'a@gmail.com'
        msg['From'] = 'b@gmail.com'
        with self.assertRaises(ValueError):
            r.execute(Email(msg.as_string()))
